@@ -1,30 +1,42 @@
 # Copyright IBM Corp. 2021. All Rights Reserved.
 
 import base64
+import io
 import json
 import os
 import warnings
+from abc import abstractmethod
 from collections import abc
-from typing import cast, ClassVar, Mapping, Union, Optional, Any, Callable
-from urllib.parse import urljoin
+from functools import wraps
+from typing import cast, ClassVar, Mapping, Union, Optional, Any, Callable, \
+    Tuple, Protocol
+from urllib.parse import urljoin, quote
 
 import requests
-import responses
 import attr
 from attr import attrs, fields_dict, NOTHING
-from ibm_cloud_sdk_core import BaseService, ApiException, DetailedResponse
-from ibm_cloud_sdk_core.authenticators import IAMAuthenticator
+from ibm_cloud_sdk_core import BaseService, DetailedResponse
+from ibm_cloud_sdk_core.authenticators import Authenticator, IAMAuthenticator, CloudPakForDataAuthenticator
 import ibm_boto3
 from ibm_botocore.client import Config
 
 from .client_errors import JsonParsingError, MissingValueError, \
     OfCpdPathError, NoWmlInstanceError, WmlServiceNameUnknownTypeError, \
-    WmlServiceNameNoPrefixError, WmlServiceCNameNotValidError
+    WmlServiceNameNoPrefixError, WmlServiceCNameNotValidError, \
+    PublicCloudOnlyError, FilesResultsNotSupportedError
 from .cpd_paths import CpdScope
 from .utils import validate_type, \
     get_storage_config_field, get_scope_response_field, get_credentials_field
 
 
+def public_cloud_only(fn: Callable) -> Callable:
+    @wraps(fn)
+    def inner(self: 'WSPipelines', *args, **kwargs):
+        if not self.is_public:
+            raise PublicCloudOnlyError("Function", fn.__name__)
+        return fn(self, *args, **kwargs)
+    return inner
+
 class WSPipelines(BaseService):
     """Watson Studio Pipelines client
 
@@ -77,20 +89,22 @@ class WSPipelines(BaseService):
                 raise MissingValueError('APIKEY')
         validate_type(apikey, "apikey", str)
 
-        authenticator = self._get_authenticator(apikey, url)
+        authenticator, is_public = self._get_authenticator(apikey, url)
 
         if service_name is None:
             service_name = self.DEFAULT_SERVICE_NAME
 
         super().__init__(
             service_url=url,
-            authenticator=authenticator
+            authenticator=authenticator,
+            disable_ssl_verification=not is_public,
         )
         self.authenticator = authenticator
         self.apikey = apikey
         self.configure_service(service_name)
+        self.is_public = is_public
 
-    def _get_iam_url(self, url: str) -> str:
+    def _get_iam_url(self, url: str) -> Tuple[str, bool]:
         validate_type(url, "url", str)
         if not url.startswith("https://"):
             warnings.warn("'url' doesn't start with https")
@@ -100,12 +114,22 @@ class WSPipelines(BaseService):
             "https://api.dataplatform.test.cloud.ibm.com": "https://iam.cloud.ibm.com/identity/token",
             "https://api.dataplatform.cloud.ibm.com": "https://iam.cloud.ibm.com/identity/token"
         }
-        return url_to_iam_url.get(url, "https://iam.cloud.ibm.com/identity/token")
+        is_public = True
+        iam_url = url_to_iam_url.get(url)
+        if iam_url is None:
+            # assume it's CPD
+            is_public = False
+            iam_url = url + "/icp4d-api/v1/authorize"
+        return iam_url, is_public
 
-    def _get_authenticator(self, apikey: str, url: str) -> IAMAuthenticator:
+    def _get_authenticator(self, apikey: str, url: str) -> Tuple[Authenticator, bool]:
         validate_type(apikey, "api_key", str)
-        iam_url = self._get_iam_url(url)
-        return IAMAuthenticator(apikey, url = iam_url)
+        iam_url, is_public = self._get_iam_url(url)
+        if is_public:
+            auth = IAMAuthenticator(apikey, url=iam_url)
+        else:
+            auth = CloudPakForDataAuthenticator("admin", url=iam_url, apikey=apikey, disable_ssl_verification=True)
+        return auth, is_public
 
     @classmethod
     def get_output_artifact_map(cls) -> Mapping[str, str]:
@@ -132,24 +156,27 @@ class WSPipelines(BaseService):
         output_artifacts = cast(Mapping[str, str], output_artifacts)
         return output_artifacts
 
-    def get_project(self, scope_id: str) -> dict:
+    def get_project(self, scope_id: str, *, context: Optional[str] = None) -> dict:
         """Get project of given ID."""
         uri = urljoin("/v2/projects/", scope_id)
-        scope = self._get_scope_from_uri(uri)
+        scope = self._get_scope_from_uri(uri, context=context)
         return scope
 
-    def get_space(self, scope_id: str) -> dict:
+    def get_space(self, scope_id: str, *, context: Optional[str] = None) -> dict:
         """Get space of given ID."""
         uri = urljoin("/v2/spaces/", scope_id)
-        scope = self._get_scope_from_uri(uri)
+        scope = self._get_scope_from_uri(uri, context=context)
         return scope
 
-    def _get_scope_from_uri(self, uri: str):
+    def _get_scope_from_uri(self, uri: str, *, context: Optional[str] = None):
         headers = {
             "Accept": "application/json",
         }
+        params = {}
+        if context is not None:
+            params["context"] = context
 
-        scope_request = self.prepare_request('GET', uri, headers = headers)
+        scope_request = self.prepare_request('GET', uri, headers=headers, params=params)
         # BaseService has some type signature problems here
         scope_request = cast(requests.Request, scope_request)
 
@@ -185,9 +212,13 @@ class WSPipelines(BaseService):
         cpd_scope: Optional[Union[str, CpdScope]] = None
     ) -> dict:
         """Get scope given its CPDPath."""
-        cpd_scope = self._get_as_scope(cpd_scope)
+        cpd_scope = self.get_scope_cpdpath(cpd_scope)
 
-        scope_type_map: Mapping[str, Callable[[str], dict]] = {
+        class ScopeGetter(Protocol):
+            @abstractmethod
+            def __call__(self, scope_id: str, *, context: Optional[str] = None) -> dict: ...
+
+        scope_type_map: Mapping[str, ScopeGetter] = {
             'projects': self.get_project,
             'spaces': self.get_space,
         }
@@ -197,7 +228,12 @@ class WSPipelines(BaseService):
             li = ', '.join(scope_type_map.keys())
             msg = "Handling scopes other than {} is not supported yet!".format(li)
             raise NotImplementedError(msg)
-        scope = scope_getter(cpd_scope.scope_id)
+
+        ctx = cpd_scope.context
+        if ctx == '':
+            ctx = None
+
+        scope = scope_getter(cpd_scope.scope_id, context=ctx)
         return scope
 
     @classmethod
@@ -208,13 +244,26 @@ class WSPipelines(BaseService):
         props = get_scope_response_field(scope_response, 'entity.storage.properties', dict)
         return props
 
+    @classmethod
+    def _extract_storage_guid(
+            cls,
+            scope_response: dict
+    ) -> str:
+        guid = get_scope_response_field(scope_response, 'entity.storage.guid', str)
+        return guid
+
+    @public_cloud_only
     def get_wml_credentials(
         self,
         cpd_scope: Optional[Union[str, CpdScope]] = None
     ) -> dict:
-        """Get WML credentials given scope's CPDPath."""
-        cpd_scope = self._get_as_scope(cpd_scope)
+        """Get WML credentials given scope's CPDPath.
 
+        Note: this is a public-cloud-only feature. For CPD, only the address
+        and API key are needed, no credentials."""
+        # make sure cpd_scope is not-None, as _extract_wml_creds_from_scope_response
+        # needs it that way
+        cpd_scope = self.get_scope_cpdpath(cpd_scope)
         scope_response = self.get_scope(cpd_scope)
 
         wml_credentials = self._extract_wml_creds_from_scope_response(
@@ -290,13 +339,15 @@ class WSPipelines(BaseService):
         else:
             raise WmlServiceCNameNotValidError(crn, cname, ['bluemix', 'staging'])
 
+    @public_cloud_only
     def get_storage_credentials(
         self,
         cpd_scope: Optional[Union[str, CpdScope]] = None
     ) -> dict:
-        """Get storage credentials given scope's CPDPath."""
-        cpd_scope = self._get_as_scope(cpd_scope)
+        """Get storage credentials given scope's CPDPath.
 
+        Note: this is a public-cloud-only feature. For CPD, only the address
+        and API key are needed, no credentials."""
         scope_response = self.get_scope(cpd_scope)
         props = self._extract_storage_properties(scope_response)
 
@@ -304,10 +355,19 @@ class WSPipelines(BaseService):
         return cos_credentials.to_dict()
 
     @classmethod
-    def _get_as_scope(
+    def get_scope_cpdpath(
         cls,
-        cpd_scope: Optional[Union[str, CpdScope]]
+        cpd_scope: Optional[Union[str, CpdScope]] = None
     ) -> CpdScope:
+        """Get the scope as CpdScope.
+
+         The operation performed depends on the data type passed:
+         * given ``None``, the default scope will be retrieved from environmental variable
+         * given a string, it will be parsed to a ``CpdScope``
+         * given a ``CpdScope``, it will be returned as-is (i.e. it's a no-op)
+
+         Mostly useful with zero arguments (to retrieve the default scope)
+         or when handling ``Union[str, CpdScope]``."""
         # if cpd_scope is None --- get it from env-var
         if cpd_scope is None:
             cpd_scope = os.environ.get('OF_CPD_SCOPE', None)
@@ -327,50 +387,33 @@ class WSPipelines(BaseService):
 
     def store_results(
         self,
-        outputs: Mapping[str, Any] # output name -> value
+        outputs: Mapping[str, Any], # output name -> value
     ) -> DetailedResponse:
         """Store notebook's results."""
         validate_type(outputs, "outputs", abc.Mapping)
         for key, value in outputs.items():
             validate_type(key, f"outputs[...]", str)
 
-        cpd_scope = os.environ.get('OF_CPD_SCOPE', None)
-        if cpd_scope is None:
-            raise MissingValueError("OF_CPD_SCOPE")
-        try:
-            cpd_scope = CpdScope.from_string(cpd_scope)
-        except Exception as ex:
-            raise OfCpdPathError(cpd_scope, reason = ex)
-        validate_type(cpd_scope, "OF_CPD_SCOPE", CpdScope)
-
+        cpd_scope = self.get_scope_cpdpath() # needed for CPD variant anyway
         scope = self.get_scope(cpd_scope)
-        props = self._extract_storage_properties(scope)
 
-        cos_config = StorageConfig.from_storage_properties(props)
-        cos_client = CosClient(self, cos_config)
+        storage_client: StorageClient
+
+        if self.is_public:
+            props = self._extract_storage_properties(scope)
+            cos_config = StorageConfig.from_storage_properties(props)
+            storage_client = CosClient(self, cos_config)
+        else:
+            guid = self._extract_storage_guid(scope)
+            storage_client = CamsClient(self, cpd_scope, guid)
 
         output_artifacts = self.get_output_artifact_map()
 
         response = None
         for output_name, output_value in outputs.items():
             result_key = output_artifacts[output_name]
-            response = cos_client.store_result(result_key, output_value)
-            if response.status not in range(200, 300):
-                raise ApiException(response.status, http_response=response)
+            response = storage_client.store_result(output_name, result_key, output_value)
 
-        if response is not None:
-            response.url = urljoin(cos_config.endpoint, cos_config.bucket_name)
-            response.headers = {}
-            response = DetailedResponse(response = response)
-        else:
-            response = DetailedResponse(
-                response=responses.Response(
-                    'PUT',
-                    url = urljoin(cos_config.endpoint, cos_config.bucket_name),
-                    status = 200
-                ),
-                status_code = 200
-            )
         return response
 
 @attrs(auto_attribs=True, kw_only=True, frozen=True)
@@ -465,6 +508,10 @@ class StorageConfig:
         validate_type(auth_endpoint, "auth_endpoint", str)
         return auth_endpoint
 
+class StorageClient(Protocol):
+    @abstractmethod
+    def store_result(self, output_name: str, output_key: str, value: Any) -> DetailedResponse: ...
+
 class CosClient:
     def __init__(
         self,
@@ -485,14 +532,80 @@ class CosClient:
             endpoint_url=config.endpoint
         )
 
-    def store_result(self, output_key: str, value: Any) -> responses.Response:
+    def store_result(self, output_name: str, output_key: str, value: Any) -> DetailedResponse:
+        validate_type(output_name, "output_name", str)
         validate_type(output_key, "output_key", str)
-        response = self.cos.Object(self.config.bucket_name, output_key).put(
-            Body=str(value)
+
+        if isinstance(value, io.TextIOBase):
+            # not supported yet
+            raise FilesResultsNotSupportedError(output_name)
+        else:
+            value = str(value)
+
+        cos_response = self.cos.Object(self.config.bucket_name, output_key).put(
+            Body=value
         )
-        return responses.Response(
+        response = requests.Response()
+        response.request = requests.Request(
             method = 'PUT',
             url = urljoin(urljoin(self.config.endpoint, self.config.bucket_name), output_key),
-            status = response['ResponseMetadata']['HTTPStatusCode'],
-            headers = response['ResponseMetadata']['HTTPHeaders']
+            headers = {},
         )
+        response.status_code = cos_response['ResponseMetadata']['HTTPStatusCode']
+        response.headers = cos_response['ResponseMetadata']['HTTPHeaders']
+        return DetailedResponse(
+            response=response
+        )
+
+class CamsClient:
+    def __init__(
+        self,
+        cpd_orchestration: WSPipelines,
+        scope: CpdScope,
+        guid: str,
+    ):
+        validate_type(cpd_orchestration, "cpd_orchestration", WSPipelines)
+        validate_type(scope, "scope", CpdScope)
+        validate_type(guid, "guid", str)
+
+        self.cpd_orchestration = cpd_orchestration
+        self.scope = scope
+        self.guid = guid
+
+    def store_result(self, output_name: str, output_key: str, value: Any) -> DetailedResponse:
+        validate_type(output_name, "output_name", str)
+        validate_type(output_key, "output_key", str)
+
+        if isinstance(value, io.TextIOBase):
+            # not supported yet
+            raise FilesResultsNotSupportedError(output_name)
+        else:
+            value = str(value)
+
+        scope_key = self.scope.scope_type
+        if scope_key.endswith('s'):
+            scope_key = scope_key[:-1]
+        scope_key += '_id'
+
+        headers = {
+            "Accept": "application/json",
+        }
+        params = {
+            scope_key: self.scope.scope_id,
+        }
+
+        if self.scope.context is not None and self.scope.context != '':
+            params["context"] = self.scope.context
+
+        asset_uri_prefix = '/v2/asset_files/'
+        asset_file_uri = quote(output_key, safe='')
+        uri = urljoin(asset_uri_prefix, asset_file_uri)
+
+        files = {
+            "file": (output_key.split('/')[-1], value)
+        }
+
+        req = self.cpd_orchestration.prepare_request('PUT', uri, headers=headers, params=params, files=files)
+        req = cast(requests.Request, req)
+        res = self.cpd_orchestration.send(req)
+        return res
\ No newline at end of file
